//+------------------------------------------------------------------+
//|                                                     OLAPDEMO.mq5 |
//|                                 Copyright © 2016-2019, Marketeer |
//|                          https://www.mql5.com/en/users/marketeer |
//|               Online Analytical Processing of trading hypercubes |
//|                            https://www.mql5.com/ru/articles/6602 |
//|                            https://www.mql5.com/ru/articles/6603 |
//+------------------------------------------------------------------+

#property copyright "Copyright © 2016-2019, Marketeer"
#property link      "https://www.mql5.com/ru/articles/6602"
#property version   "1.0"
#property description "Trade History Analysis: Non-trading Expert Adviser demonstrating OLAP classes"

#include <OLAP/OLAPcube.mqh>
#include <OLAP/HTMLcube.mqh>
#include <OLAP/CSVcube.mqh>


input GroupSettings OLAP_Settings; // O L A P     S E T T I N G S

sinput string X = "————— X axis —————";
input SELECTORS SelectorX = SELECTOR_SYMBOL;
input TRADE_RECORD_FIELDS FieldX = FIELD_NONE /* field does matter only for some selectors */;

sinput string Y = "————— Y axis —————";
input SELECTORS SelectorY = SELECTOR_NONE;
input TRADE_RECORD_FIELDS FieldY = FIELD_NONE;

sinput string Z = "————— Z axis —————";
input SELECTORS SelectorZ = SELECTOR_NONE;
input TRADE_RECORD_FIELDS FieldZ = FIELD_NONE;

sinput string F = "————— Filter —————";
input SELECTORS Filter1 = SELECTOR_NONE;
input TRADE_RECORD_FIELDS Filter1Field = FIELD_NONE;
input double Filter1value1 = 0;
input double Filter1value2 = 0;

sinput string A = "————— Aggregator —————";
input AGGREGATORS AggregatorType = AGGREGATOR_SUM;
input TRADE_RECORD_FIELDS AggregatorField = FIELD_PROFIT_AMOUNT;
input SORT_BY SortBy = SORT_BY_NONE;

int selectorCount;
SELECTORS selectorArray[4];
TRADE_RECORD_FIELDS selectorField[4];

#define PRT(A) Print(#A, " ", (A))

int OnInit()
{
  selectorCount = (SelectorX != SELECTOR_NONE) + (SelectorY != SELECTOR_NONE) + (SelectorZ != SELECTOR_NONE);
  selectorArray[0] = SelectorX;
  selectorArray[1] = SelectorY;
  selectorArray[2] = SelectorZ;
  selectorArray[3] = Filter1;
  selectorField[0] = FieldX;
  selectorField[1] = FieldY;
  selectorField[2] = FieldZ;
  selectorField[3] = Filter1Field;

  EventSetTimer(1);
  return(INIT_SUCCEEDED);
}

void OnTimer()
{
  process();
  EventKillTimer();
}

class DaysRangeSelector: public DateTimeSelector<TRADE_RECORD_FIELDS>
{
  public:
    DaysRangeSelector(const int n): DateTimeSelector<TRADE_RECORD_FIELDS>(FIELD_DURATION, n)
    {
      _typename = typename(this);
    }
    
    virtual bool select(const Record *r, int &index) const override
    {
      double d = r.get(selector);
      int days = (int)(d / (60 * 60 * 24));
      index = MathMin(days, granularity - 1);
      return true;
    }
    
    virtual string getLabel(const int index) const override
    {
      return index < granularity - 1 ? ((index < 10 ? " ": "") + (string)index + "D") : ((string)index + "D+");
    }
};

Selector<TRADE_RECORD_FIELDS> *createSelector(int i)
{
  switch(selectorArray[i])
  {
    case SELECTOR_TYPE:
      return new TypeSelector();
    case SELECTOR_SYMBOL:
      return new SymbolSelector();
    case SELECTOR_SERIAL:
      return new SerialNumberSelector();
    case SELECTOR_MAGIC:
      return new MagicSelector();
    case SELECTOR_PROFITABLE:
      return new ProfitableSelector();
    case SELECTOR_DURATION:
      return new DaysRangeSelector(15); // up to 14 days
    case SELECTOR_WEEKDAY:
      return selectorField[i] != FIELD_NONE ? new WeekDaySelector(selectorField[i]) : NULL;
    case SELECTOR_DAYHOUR:
      return selectorField[i] != FIELD_NONE ? new DayHourSelector(selectorField[i]) : NULL;
    case SELECTOR_HOURMINUTE:
      return selectorField[i] != FIELD_NONE ? new DayHourSelector(selectorField[i]) : NULL;
    case SELECTOR_SCALAR:
      return selectorField[i] != FIELD_NONE ? new TradeSelector(selectorField[i]) : NULL;
    case SELECTOR_QUANTS:
      return selectorField[i] != FIELD_NONE ? new QuantizationSelector(selectorField[i]) : NULL;
  }
  return NULL;
}

void process()
{
  if(selectorCount == 0)
  {
    Print("No selectors. Setup at least one of them.");
    return;
  }
  HistoryDataAdapter<CustomTradeRecord> history;
  HTMLReportAdapter<CustomTradeRecord> report;
  CSVReportAdapter<CustomTradeRecord> external;
  
  DataAdapter *adapter = &history;
  
  if(ReportFile != "")
  {
    if(StringFind(ReportFile, ".htm") > 0 && report.load(ReportFile))
    {
      adapter = &report;
    }
    else
    if(StringFind(ReportFile, ".csv") > 0 && external.load(ReportFile))
    {
      adapter = &external;
    }
    else
    {
      Print("Unknown file format: ", ReportFile);
      return;
    }
  }
  else
  {
    Print("Analyzing account history");
  }
  
  Analyst<TRADE_RECORD_FIELDS> *analyst;
  
  Selector<TRADE_RECORD_FIELDS> *selectors[];
  ArrayResize(selectors, selectorCount);
  
  for(int i = 0; i < selectorCount; i++)
  {
    selectors[i] = createSelector(i);
    if(selectors[i] == NULL)
    {
      Print("Selector ", i, " is empty. Setup selectors successively (don't leave a hole in-between), specify a field when required");
      return;
    }
  }
  Filter<TRADE_RECORD_FIELDS> *filters[];
  if(Filter1 != SELECTOR_NONE)
  {
    ArrayResize(filters, 1);
    Selector<TRADE_RECORD_FIELDS> *filterSelector = createSelector(3);
    if(Filter1value1 != Filter1value2)
    {
      filters[0] = new FilterRange<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1, Filter1value2);
    }
    else
    {
      filters[0] = new Filter<TRADE_RECORD_FIELDS>(filterSelector, Filter1value1);
    }
  }
  
  Aggregator<TRADE_RECORD_FIELDS> *aggregator;
  
  // MQL does not support a 'class info' metaclass.
  // Otherwise we could use an array of classes instead of the switch
  switch(AggregatorType)
  {
    case AGGREGATOR_SUM:
      aggregator = new SumAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_AVERAGE:
      aggregator = new AverageAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_MAX:
      aggregator = new MaxAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_MIN:
      aggregator = new MinAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_COUNT:
      aggregator = new CountAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_PROFITFACTOR:
      aggregator = new ProfitFactorAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_PROGRESSIVE:
      aggregator = new ProgressiveTotalAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
    case AGGREGATOR_IDENTITY:
      aggregator = new IdentityAggregator<TRADE_RECORD_FIELDS>(AggregatorField, selectors, filters);
      break;
  }
  
  LogDisplay display(10, 3);
  
  analyst = new Analyst<TRADE_RECORD_FIELDS>(adapter, aggregator, display);
  
  analyst.acquireData();
  
  Print("Symbol number: ", TradeRecord::getSymbolCount());
  for(int i = 0; i < TradeRecord::getSymbolCount(); i++)
  {
    Print(i, "] ", TradeRecord::getSymbol(i));
  }

  Print("Magic number: ", TradeRecord::getMagicCount());
  for(int i = 0; i < TradeRecord::getMagicCount(); i++)
  {
    Print(i, "] ", TradeRecord::getMagic(i));
  }
  
  Print("Filters: ", aggregator.getFilterTitles());
  
  Print("Selectors: ", selectorCount);
  
  analyst.build();
  analyst.display(SortBy);

  /* test getVector
  int consts[];
  ArrayResize(consts, selectorCount);
  ArrayInitialize(consts, 0);
  PairArray result;
  aggregator.getVector(0, consts, result, SortBy);
  ArrayPrint(result.array);
  */

  delete analyst;
  delete aggregator;
  for(int i = 0; i < selectorCount; i++)
  {
    delete selectors[i];
  }
  for(int i = 0; i < ArraySize(filters); i++)
  {
    delete filters[i].getSelector();
    delete filters[i];
  }
}

void OnTick()
{
}
